//
// Reminder.cs: A tomboy plugin that displays a reminder popup at specified dates.
//
// Copyright 2005,  Raphaël Slinckx <raphael@slinckx.net>
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
// 
// http://www.apache.org/licenses/LICENSE-2.0
// 
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.
//

using System;
using System.IO;
using Gtk;
using System.Text.RegularExpressions;
using System.Threading;
using System.Collections;
using MagicString;

namespace Tomboy {
	public class Reminder : NotePlugin
	{
		//The reminder date expression, currently (in a non-langage specific fashion) it is
		// !date
		//or
		// remind date | alert: next monday | rappel mardi prochain
		//anywhere once in a line
		static string[] REMINDER_PATTERNS = {
			".*\\!",
			"(remind|rappel|alert)\\W+"
		};
		static Regex REMINDER = new Regex ("^(("+REMINDER_PATTERNS[0]+")|("+REMINDER_PATTERNS[1]+"))(?<date>.*)$", RegexOptions.IgnoreCase | RegexOptions.Compiled | RegexOptions.Multiline);
		static MagicDate magicDate = new MagicDate();
		static ArrayList pendingEvents;
		
		//How much time after the notification popup disappears
		static int NOTIFICATION_TIMEOUT = 5000;
		
		protected override void Initialize ()
		{
			pendingEvents = new ArrayList();
			
			//Console.WriteLine("Reminder.cs:Reminder Initialize");
			foreach(Note note in Manager.Notes) {
				string[] lines = note.TextContent.Split('\n');
				foreach(string line in lines) {
					SetupTimer(note, line);
					
				}
			}
		}
		
		protected override void Shutdown () {}
		
		protected override void OnNoteOpened () 
		{
			Note.Window.DeleteEvent += OnNoteClosed;
		}
		
		[GLib.ConnectBefore]
		void OnNoteClosed (object sender, DeleteEventArgs args) 
		{
			RemoveEveryNote(Note);
			string[] lines = Note.TextContent.Split('\n');
			foreach(string line in lines)
				SetupTimer(Note, line);
			
			Note.Window.DeleteEvent -= OnNoteClosed;
		}
		
		private void RemoveEveryNote(Note note)
		{
			ArrayList toRemove = new ArrayList();
			foreach(ReminderEvent ev in pendingEvents) {
				if(ev.Note.Equals(note)) {
					Console.WriteLine("Reminder.cs:Removing deleted reminder: "+ev.Date);
					toRemove.Add(ev);
				}
			}
			
			foreach(ReminderEvent ev in toRemove)
				pendingEvents.Remove(ev);
		}

		private void SetupTimer (Note note, TextIter pos)
		{
			TextIter start = pos.Copy();
			start.BackwardChars(start.VisibleLineOffset);
			TextIter end = pos.Copy();
			end.ForwardToLineEnd();
			
			SetupTimer(Note, start.GetText(end));
		}
		
		private void SetupTimer (Note note, string line)
		{
			for (Match match = REMINDER.Match (line); match.Success; match = match.NextMatch ()) {
				DateTime time;
				try {
					Console.WriteLine("Reminder.cs:Parsing date:"+match.Groups ["date"].ToString());
					time = magicDate.ParseDate(match.Groups ["date"].ToString(), note.ChangeDate);
					//Console.WriteLine("Reminder.cs:Parsing OK");
				} catch(Exception e) {
					//Console.WriteLine("Reminder.cs:Parsing Failed");
					continue;
				}
				
				
				DateTime now = DateTime.Now;
				long due = (time.Ticks - now.Ticks)/10000;
				if(time.Day == now.Day && time.Month == now.Month && time.Year == now.Year) {
					//For immediate release !
					due = 0;
				}
				
				//We have a past reminder, forget it.
				if(due < 0)
					continue;
				
				
				ReminderEvent ev = new ReminderEvent(note, time, (int) due);
				if(pendingEvents.Contains(ev))
					continue;
				
				//Console.WriteLine("Reminder.cs:Timer in : "+(due/1000)+" sec|@ "+time);
				Console.WriteLine("Reminder.cs:Timer @ "+time);
				pendingEvents.Add(ev);
				ev.Start();
			}
		}
		
		private class ReminderEvent {
			private Note note;
			private DateTime date;
			private int dueTime;
			private Gtk.Window notif;
			
			ThreadNotify showReminder;
			ThreadNotify removeReminder;
			
			public ReminderEvent(Note note, DateTime date, int dueTime) {
				this.showReminder = new ThreadNotify(new ReadyEvent(OnShowReminder));
				this.removeReminder = new ThreadNotify(new ReadyEvent(OnRemoveReminder));
				
				this.note = note;
				this.date = date;
				this.dueTime = dueTime;
			}
			
			public void Start () {
				new Timer(new TimerCallback(OnShowReminderTimer), null, dueTime, System.Threading.Timeout.Infinite );
			}
			
			private void OnShowReminderTimer (object state)
			{
				//Console.WriteLine("Reminder.cs:OnShowReminderTimer");
				showReminder.WakeupMain();
			}
			
			private void OnShowReminder ()
			{
				this.notif = new Gtk.Window(WindowType.Toplevel);
				notif.Title = "Tomboy Reminder";
				notif.Resizable = false;
				notif.SetPosition(WindowPosition.None);
				notif.Stick();
				notif.SkipTaskbarHint = true;
				notif.SetSizeRequest(350, -1);
				notif.IsFocus = false;
				notif.TypeHint = Gdk.WindowTypeHint.Splashscreen;
				notif.ModifyBg(Gtk.StateType.Normal, new Gdk.Color(247,255,157));
				
				VBox box = new VBox(false, 4);
				Label l = new Label();
				l.UseMarkup = true;
				l.Wrap = true;
				l.LineWrap = true;
				l.Justify = Gtk.Justification.Center;
				l.Markup = String.Format("<b>{0} {1}</b>", note.Title, note.ChangeDate.ToString("dd/MM/yyyy"));
				
				Label l2 = new Label ();
				l2.UseMarkup = false;
				l2.Wrap = true;
				l2.LineWrap = true;
				l2.Justify = Gtk.Justification.Center;
				l2.Text = note.TextContent;
				
				box.Add(l);
				box.Add(l2);
				notif.Add(box);
				
				notif.Realize();
				notif.Decorated = false;
				
				notif.Hide();
				notif.Move(0, 0);
				
				notif.Gravity = Gdk.Gravity.NorthWest;
				notif.Move(Gdk.Screen.Default.Width-350, 0);
				notif.ShowAll();
				
				new Timer(new TimerCallback(OnRemoveReminderTimer), null, NOTIFICATION_TIMEOUT, System.Threading.Timeout.Infinite);
			}
			
			private void OnRemoveReminderTimer (object state)
			{
				//Console.WriteLine("Reminder.cs:OnRemoveReminderTimer");
				pendingEvents.Remove(this);
				removeReminder.WakeupMain();
			}
			
			private void OnRemoveReminder ()
			{
				notif.Destroy();
			}
			
			public Note Note {
				get { return note; }
			}
			
			public DateTime Date {
				get { return date; }
			}
			
			public override bool Equals (object o) {
				ReminderEvent ev = o as ReminderEvent;
				return note.Equals(ev.Note) &&
					(date.Day == ev.Date.Day && date.Month == ev.Date.Month && date.Year == ev.Date.Year);
			}
		}
	}
}
